open Syntax
open Types
open Coercions
open Terms

type identifier =
    Identifier.identifier

type atom =
    Atom.atom

type env =
    Import.env

(* ------------------------------------------------------------------------- *)

(* Record labels are implicitly bound at the top level. *)

let ilabel : identifier -> atom =
  Import.mkglobal()

(* ------------------------------------------------------------------------- *)

(* Term variables and type variables are explicitly bound. *)

let bind env id : atom * env =
  let env = Import.bind env id in
  Import.resolve env id, env

(* ------------------------------------------------------------------------- *)

(* [ifields] turns an association list whose keys are identifiers
   (record labels) into a map whose keys are atoms. It checks that no
   label appears twice. *)

let ifields (f : 'a -> 'b) (fields : (identifier * 'a) list) : 'b Atom.AtomMap.t =
  Import.linear (List.map fst fields);
  Atom.AtomMap.of_list
    (List.map (function key, data -> ilabel key, f data) fields)

(* ------------------------------------------------------------------------- *)

(* [itype] converts a type from external syntax into internal syntax. *)

let rec itype env : Syntax.ftype -> Types.ftype = function
  | SynTyVar id ->
      TyFreeVar (Import.resolve env id)
  | SynTyArrow (domain, codomain) ->
      TyArrow (itype env domain, itype env codomain)
  | SynTyRecord fields ->
      TyRecord (ifields (itype env) fields)
  | SynTyQuantified (q, a, body) ->
      let a, env = bind env a in
      TyQuantified (q, abstract a (itype env body))

(* ------------------------------------------------------------------------- *)

(* [icoercion] converts a coercion from external syntax into internal syntax. *)

let rec icoercion env = function
  | SynCoId -> CoId
  | SynCoIntroForall -> CoIntroForall
  | SynCoElimExists -> CoElimExists
  | SynCoDistrib -> CoDistrib
  | SynCoDistribLabel l -> CoDistribLabel (ilabel l)
  | SynCoElimForall (m, typ) -> 
      let typ = itype env typ in
      CoElimForall (m, typ)
  | SynCoIntroExists (m, typ) -> 
      let typ = itype env typ in
      CoIntroExists (m, typ)
  | SynCoTrans (c1, c2) ->
      let c1 = iloccoercion env c1 in
      let c2 = iloccoercion env c2 in
      CoTrans (c1, c2)
  | SynCoArrow (c1, c2) -> 
      let c1 = iloccoercion env c1 in
      let c2 = iloccoercion env c2 in
      CoArrow (c1, c2)
  | SynCoQuantifier (q, a, c) ->
      let a, env = bind env a in
      let c = iloccoercion env c in
      CoQuantifier (q, a, c)
  | SynCoRecord fields ->
      Import.linear (List.map fst fields);
      CoRecord
        (Atom.AtomMap.of_list
           (List.map
              (function key, c -> ilabel key, iloccoercion env c)
              fields))

and iloccoercion env (loc, c) = loc, icoercion env c

(* ------------------------------------------------------------------------- *)

(* [iterm] and [ilocterm] convert a term from external syntax into
   internal syntax. *)

let rec iterm env = function

  | SynTeVar id ->
      TeVar (Import.resolve env id)

  | SynTeFun (x, domain, codomain, body) ->
      (* a non-recursive function is viewed as sugar for a recursive one *)
      let x, env = bind env x in
      let codomain =
        match codomain with None -> None | Some ty -> Some (itype env ty) in
      TeFun (x, itype env domain, codomain, ilocterm env body)

  | SynTeRecFun (f, x, domain, codomain, body) ->
      (* order matters here: [x] could hide [f] *)
      let f, env = bind env f in
      let x, env = bind env x in
      TeRecFun (f, x, itype env domain, itype env codomain, ilocterm env body)

  | SynTeApp (term1, term2) ->
      TeApp (ilocterm env term1, ilocterm env term2)

  | SynTeLet (x, term1, term2) ->
      let x, env' = bind env x in
      TeLet (x, ilocterm env term1, ilocterm env' term2)

  | SynTeTyAbs (a, term) ->
      let a, env = bind env a in
      TeTyAbs (a, ilocterm env term)

  | SynTeTyApp (term, ftype) ->
      TeTyApp (ilocterm env term, itype env ftype)

  | SynTeRecord fields ->
      TeRecord (ifields (ilocterm env) fields)

  | SynTeProj (term, label) ->
      TeProj (ilocterm env term, ilabel label)

  | SynTePack (term, a, ty) ->
      (* [a] is bound in [ty], and only there *)
      let term = ilocterm env term in
      let a, env = bind env a in
      TePack (term, abstract a (itype env ty))

  | SynTeUnpack (a, x, term1, term2) ->
      let term1 = ilocterm env term1 in
      (* [a] and [x] are bound in [term2] *)
      let a, env = bind env a in
      let x, env = bind env x in
      TeUnpack (a, x, term1, ilocterm env term2)

  | SynTeCoerce (c, term) ->
      let c = iloccoercion env c in
      let term = ilocterm env term in
      TeCoerce (c, term)

and ilocterm env (loc, t) =
  (loc, iterm env t)

(* ------------------------------------------------------------------------- *)

(* [program] converts a complete program. *)

let program p =
  let t = ilocterm Import.empty p in
  Error.signaled();
  t

