%{

open Syntax

let mkid sort (name, startp, endp) =
  Identifier.make name sort startp endp

let dummy =
  Error.dummy

let tyabs tvs t =
  snd (
    List.fold_right (fun tv t ->
      dummy, SynTeTyAbs (tv, t)
    ) tvs t
  )

%}

%token <string * Lexing.position * Lexing.position> IDENTIFIER
%token COMMA SEMI DOT COLON 
%token ARROW EQ
%token DISTRIB  PLUS MINUS SHARP
%token LPAR RPAR LBRACE RBRACE LBRACKET RBRACKET LANGLE RANGLE
%token FUN LET IN REC EXISTS FORALL PACK UNPACK
%token EOF
%start <Syntax.loc_fterm> program

%%

(* ------------------------------------------------------------------------- *)

(* Recording locations. *)

loc(X):
| x = X
    { (($startpos, $endpos), x) }

(* ------------------------------------------------------------------------- *)

(* In ocaml syntax, SEMI can be used either as a delimiter or as a
   terminator. That is, the last SEMI is optional. Here is a way of
   saying so. *)

semi(X):
| (* nothing *)
    { [] }
| xs = terminated(X, SEMI)+
| xs = separated_nonempty_list(SEMI, X)
    { xs }

(* ------------------------------------------------------------------------- *)

(* Identifiers. *)

(* Each identifier is tagged with its sort, which is known thanks
   to the context where the identifier appears. *)

term_variable:
| id = IDENTIFIER
    { mkid term_sort id }

type_variable:
| id = IDENTIFIER
    { mkid type_sort id }

label:
| id = IDENTIFIER
    { mkid label_sort id }

(* ------------------------------------------------------------------------- *)

(* Quantifiers. *)

quantifier:
| FORALL
    { QForall }
| EXISTS
    { QExists }

(* ------------------------------------------------------------------------- *)

(* Types. *)

typ0:
| a = type_variable
    { SynTyVar a }
| LBRACE fields = semi(label_typ) RBRACE
    { SynTyRecord fields }
| LPAR ty = typ RPAR
    { ty }

typ:
| ty = typ0
    { ty }
| ty1 = typ0 ARROW ty2 = typ
    { SynTyArrow (ty1, ty2) }
| q = quantifier a = type_variable DOT ty = typ
    { SynTyQuantified (q, a, ty) }

label_typ:
| label = label COLON ty = typ
    { label, ty }

(* ------------------------------------------------------------------------- *)
(* Coercions. *)

mode: 
| RANGLE
    { true }
| LANGLE
    { false }

coercion0:
| SHARP
    { SynCoId }
| PLUS FORALL
    { SynCoIntroForall }
| MINUS EXISTS
    { SynCoElimExists }
| DISTRIB
    { SynCoDistrib }
| LBRACE fields = semi(label_coercion) RBRACE
    { SynCoRecord fields }
| LPAR c = coercion RPAR
    { c }

coercion1:
| c = coercion0
    { c }
| q = quantifier a = type_variable DOT c = loc(coercion1)
    { SynCoQuantifier (q, a, c) }
| MINUS FORALL m = mode ty = typ 
    { SynCoElimForall (m, ty) }
| PLUS EXISTS m = mode ty = typ 
    { SynCoIntroExists (m, ty) }
| DISTRIB DOT l = label
    { SynCoDistribLabel l }
| c1 = loc(coercion0) ARROW c2 = loc(coercion1) 
    { SynCoArrow (c1, c2) }

coercion:
| c = coercion1
    { c }
| c1 = loc(coercion1) SEMI c2 = loc(coercion)
    {  SynCoTrans (c1, c2) }

label_coercion:
| label = label COLON c = loc(coercion1)
    { label, c }


(* ------------------------------------------------------------------------- *)
(* Terms. *)

term0:
| x = term_variable
    { SynTeVar x }
| LBRACE fields = semi(label_term) RBRACE
    { SynTeRecord fields }
| t = loc(term0) DOT label = label
    { SynTeProj (t, label) }
| LPAR t = term RPAR
    { t }

term1:
| t = term0
    { t }
| t1 = loc(term1) t2 = loc(term0)
    { SynTeApp (t1, t2) }
| t = loc(term1) LBRACKET ty = typ RBRACKET
    { SynTeTyApp (t, ty) }

typeannot: 
| 
    { None }
| COLON codomain = typ 
    { Some codomain }

term:
| t = term1
    { t }
| FUN LPAR x = term_variable COLON domain = typ RPAR 
  codomain = typeannot EQ t = loc(term)
    { SynTeFun (x, domain, codomain, t) }
| FUN f = term_variable LPAR x = term_variable COLON domain = typ RPAR
  COLON codomain = typ EQ t = loc(term)
    { SynTeRecFun (f, x, domain, codomain, t) }
| FUN tvs = type_variable+ ARROW t = loc(term)
    { tyabs tvs t }
| LET x = term_variable EQ t1 = loc(term) IN t2 = loc(term)
    { SynTeLet (x, t1, t2) }
| LET f = term_variable LPAR x = term_variable COLON domain = typ RPAR
  codomain = typeannot EQ t1 = loc(term) IN t2 = loc(term)
    { SynTeLet (f, (($startpos(f), $endpos(t1)),
                    SynTeFun (x, domain, codomain, t1)), t2) }
| LET REC f = term_variable LPAR x = term_variable COLON domain = typ RPAR
  COLON codomain = typ EQ t1 = loc(term) IN t2 = loc(term)
    { SynTeLet (f, (($startpos(f), $endpos(t1)), SynTeRecFun (f, x, domain, codomain, t1)), t2) }
| PACK t = loc(term) COLON EXISTS a = type_variable DOT ty = typ
    { SynTePack (t, a, ty) }
| UNPACK a = type_variable COMMA x = term_variable EQ t1 = loc(term) IN t2 = loc(term)
    { SynTeUnpack (a, x, t1, t2) }
| LANGLE c = loc(coercion) RANGLE t = loc(term)
    { SynTeCoerce (c, t) }

label_term:
| label = label EQ t = loc(term)
    { label, t }

(* ------------------------------------------------------------------------- *)

(* Programs. *)

program:
| t = loc(term) EOF
    { t }

