open Atom
open Types
open Coercions

(* There is no polymorphic recursion: recursive functions have no type
   arguments. Ideally, this should be fixed: a recursive function
   should be able to accept an arbitrary number of type arguments. *)

(* Term variables are represented by atoms. Terms are constructed by
   [Internalize], which guarantees that all atoms are unique and that
   there are no dangling free atoms. *)

(* Terms are annotated with source code locations, which are used when
   reporting type errors. *)

type fterm =
  | TeVar of atom
      (* x *)
  | TeRecFun of atom * atom * ftype * ftype * loc_fterm
      (* fun f (x : T) : T = t *)
  | TeFun of atom * ftype * ftype option * loc_fterm
      (* fun (x : T) : T = t *)
  | TeApp of loc_fterm * loc_fterm
      (* t t *)
  | TeLet of atom * loc_fterm * loc_fterm
      (* let x = t in t *)
  | TeTyAbs of atom * loc_fterm
      (* fun a -> t *)
  | TeTyApp of loc_fterm * ftype
      (* t [T] *)
  | TeRecord of loc_fterm AtomMap.t
      (* { L = t; ...; L = t } *)
  | TeProj of loc_fterm * atom
      (* t.L *)
  | TePack of loc_fterm * ftype_context
      (* pack t : exists a. T *)
  | TeUnpack of atom * atom * loc_fterm * loc_fterm
      (* unpack x, a = t in t *)
  | TeCoerce of loc_coercion * loc_fterm
      (* <c> t *)

and loc_fterm =
    Error.location * fterm

(* ------------------------------------------------------------------------- *)

(* Helpers for constructing terms with dummy locations  *)

val var: atom -> loc_fterm
val recabs: atom -> atom -> ftype -> ftype -> loc_fterm -> loc_fterm
val abs: atom -> ftype -> loc_fterm -> loc_fterm
val app: loc_fterm -> loc_fterm -> loc_fterm
val def: (* let *) atom -> loc_fterm -> loc_fterm -> loc_fterm
val tyabs: atom -> loc_fterm -> loc_fterm
val tyapp: loc_fterm -> ftype -> loc_fterm
val record: loc_fterm AtomMap.t -> loc_fterm
val proj: loc_fterm -> atom -> loc_fterm
val pack: loc_fterm -> ftype_context -> loc_fterm
val unpack: atom -> atom -> loc_fterm -> loc_fterm -> loc_fterm


(* ------------------------------------------------------------------------- *)

(* Determining a term's free term variables. *)

val fv: fterm -> AtomSet.t

